from joblib import register_parallel_backend
from joblib_modal._modal import ModalBackend

register_parallel_backend("modal", ModalBackend)

__version__ = "0.1.0"

__all__ = ["ModalBackend", "__version__"]



import joblib_modal  # noqa
from joblib import parallel_config, Parallel, delayed
from sklearn.datasets import make_classification
from sklearn.ensemble import HistGradientBoostingClassifier
from sklearn.model_selection import RandomizedSearchCV


def test_modal_backend():
    with parallel_config(backend="modal", modal_output=True):
        out = Parallel(n_jobs=2)(delayed(lambda x: x * x)(i) for i in range(10))
        assert out == [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]


def get_model(use_modal: bool):
    X, y = make_classification(
        n_samples=1000, n_features=20, n_classes=2, random_state=42
    )

    model = RandomizedSearchCV(
        HistGradientBoostingClassifier(),
        param_distributions={
            "max_depth": [3, 5, 7],
            "learning_rate": [0.01, 0.05, 0.1],
            "max_leaf_nodes": [10, 20, 30],
            "max_bins": [10, 100, 255],
        },
        n_iter=20,
        cv=5,
        random_state=42,
    )

    if use_modal:
        import sklearn
        import numpy
        import joblib
        import modal
        import scipy

        image = (
            modal.Image.debian_slim()
            .pip_install(f"scikit-learn=={sklearn.__version__}")
            .pip_install(f"numpy=={numpy.__version__}")
            .pip_install(f"joblib=={joblib.__version__}")
            .pip_install(f"scipy=={scipy.__version__}")
        )

        with parallel_config(
            backend="modal",
            n_jobs=-1,
            name="test-joblib",
            image=image,
            modal_output=True,
        ):
            model = model.fit(X, y)
    else:
        model = model.fit(X, y)

    return model


def test_modal_with_image():
    modal_model = get_model(use_modal=True)
    native_model = get_model(use_modal=False)

    assert modal_model.best_params_ == native_model.best_params_



from joblib_modal import enable_joblib_modal  # noqa
from sklearn.model_selection import GridSearchCV
from joblib import parallel_config

...
est = GridSearchCV(...)
with parallel_config(backend="modal", inner_max_num_threads=2):
    est.fit(X, y)



import uuid
from joblib._parallel_backends import (
    FallbackToBackend,
    SequentialBackend,
    ThreadingBackend,
)
from joblib import register_parallel_backend
from joblib._utils import _TracebackCapturingWrapper
import modal


def executor(func, *args, **kwargs):
    return func(*args, **kwargs)


class ModalBackend(ThreadingBackend):
    uses_threads = True
    supports_sharedmem = False

    def __init__(self, *args, name=None, image=None, modal_output=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.image = image
        self.modal_output = modal_output

    def configure(self, n_jobs=1, parallel=None, **_):
        """Build a process or thread pool and return the number of workers"""
        n_jobs = self.effective_n_jobs(n_jobs)

        if n_jobs == 1:
            # Avoid unnecessary overhead and use sequential backend instead.
            raise FallbackToBackend(SequentialBackend(nesting_level=self.nesting_level))

        self.parallel = parallel
        self._n_jobs = n_jobs

        if self.image is None:
            image = modal.Image.debian_slim().pip_install("joblib")
        else:
            image = self.image

        name = self.name or f"modal-joblib-{uuid.uuid4()}"
        self.modal_app = modal.App(name, image=image)

        self.modal_executor = self.modal_app.function()(executor)

        if self.modal_output:
            self.output_ctx = modal.enable_output()
            self.output_ctx.__enter__()
        self.run_ctx = self.modal_app.run()
        self.run_ctx.__enter__()

        return n_jobs

    def effective_n_jobs(self, n_jobs):
        """Determine the number of jobs which are going to run in parallel"""
        if n_jobs == 0:
            raise ValueError("n_jobs == 0 in Parallel has no meaning")
        if n_jobs < 0:
            return 1000
        return n_jobs

    def apply_async(self, func, callback=None):
        """Schedule a func to be run"""
        # Here, we need a wrapper to avoid crashes on KeyboardInterruptErrors.
        # We also call the callback on error, to make sure the pool does not
        # wait on crashed jobs.
        return self._get_pool().apply_async(
            _TracebackCapturingWrapper(self.modal_executor.remote),
            (),
            {"func": func},
            callback=callback,
            error_callback=callback,
        )

    def terminate(self):
        if self.modal_output:
            self.output_ctx.__exit__(None, None, None)
        self.run_ctx.__exit__(None, None, None)
        super().terminate()


register_parallel_backend("modal", ModalBackend)
